#include <TinyScreen.h>
#include "SoundStream.hpp"

unsigned char stream_buffer[SOUND_BUFFER_SIZE];
int sample_index;
int write_buffer;
bool request;

bool tcIsSyncing(void)
{
	return TC5->COUNT16.STATUS.reg & TC_STATUS_SYNCBUSY;
}

void tcReset(void)
{
	// Reset TCx
	TC5->COUNT16.CTRLA.reg = TC_CTRLA_SWRST;
	while(tcIsSyncing());
	while(TC5->COUNT16.CTRLA.bit.SWRST);
}

void InitializeSoundStream(void)
{
	memset(stream_buffer, 0x80, SOUND_BUFFER_SIZE);
	sample_index = 0;
	write_buffer = 0;
	request = false;
	analogWrite(A0, 0);
	// Enable GCLK for TCC2 and TC5 (timer counter input clock)
	GCLK->CLKCTRL.reg = static_cast<unsigned short>((GCLK_CLKCTRL_CLKEN | GCLK_CLKCTRL_GEN_GCLK0 | GCLK_CLKCTRL_ID(GCM_TC4_TC5)));
	while(GCLK->STATUS.bit.SYNCBUSY);
	tcReset();
	// Set Timer counter Mode to 16 bits
	TC5->COUNT16.CTRLA.reg |= TC_CTRLA_MODE_COUNT16;
	// Set TC5 mode as match frequency
	TC5->COUNT16.CTRLA.reg |= TC_CTRLA_WAVEGEN_MFRQ;
	TC5->COUNT16.CTRLA.reg |= TC_CTRLA_PRESCALER_DIV1;
	TC5->COUNT16.CC[0].reg = static_cast<unsigned short>((SystemCoreClock / SOUND_FREQUENCY - 1));
	while(tcIsSyncing());
	// Configure interrupt request
	NVIC_DisableIRQ(TC5_IRQn);
	NVIC_ClearPendingIRQ(TC5_IRQn);
	NVIC_SetPriority(TC5_IRQn, 0);
	NVIC_EnableIRQ(TC5_IRQn);
	// Enable the TC5 interrupt request
	TC5->COUNT16.INTENSET.bit.MC0 = 1;
	while(tcIsSyncing());
	// Enable TC
	TC5->COUNT16.CTRLA.reg |= TC_CTRLA_ENABLE;
	while(tcIsSyncing());
}

void FinalizeSoundStream(void)
{
	// Disable TC5
	TC5->COUNT16.CTRLA.reg &= ~TC_CTRLA_ENABLE;
	while(tcIsSyncing());
	// Reset TCx
	TC5->COUNT16.CTRLA.reg = TC_CTRLA_SWRST;
	while(tcIsSyncing());
	while(TC5->COUNT16.CTRLA.bit.SWRST);
}

void UpdateSoundStream(void)
{
	int play_buffer = 0;
	if(sample_index >= SOUND_BUFFER_SIZE_HALF)
	{
		play_buffer = 1;
	}
	if(write_buffer == play_buffer)
	{
		request = true;
	}
}

int GetSoundStreamWritableSize(void)
{
	if(request == false)
	{
		return 0;
	}
	return SOUND_BUFFER_SIZE_HALF;
}

bool WriteSoundStream(const unsigned char* buffer)
{
	int play_buffer = 0;
	if(sample_index >= SOUND_BUFFER_SIZE_HALF)
	{
		play_buffer = 1;
	}
	write_buffer = 1 - play_buffer;
	memcpy(&stream_buffer[SOUND_BUFFER_SIZE_HALF * write_buffer], buffer, SOUND_BUFFER_SIZE_HALF);
	request = false;
	return true;
}

#ifdef __cplusplus
extern "C"
{
#endif
	void SoundStreamCallback(void)
	{
		while(DAC->STATUS.bit.SYNCBUSY == 1);
		DAC->DATA.reg = static_cast<unsigned short>(stream_buffer[sample_index] << 2); // 8bit->10bit
		while(DAC->STATUS.bit.SYNCBUSY == 1);
		++ sample_index;
		if(sample_index >= SOUND_BUFFER_SIZE)
		{
			sample_index = 0;
		}
		TC5->COUNT16.INTFLAG.bit.MC0 = 1;
	}
	void TC5_Handler(void) __attribute__ ((weak, alias("SoundStreamCallback")));
#ifdef __cplusplus
}
#endif
